@@ -1,13 +1,15 @@
-from typing import Optional
+import warnings
+from typing import Optional, Union
 
 import numpy as np
 import numpy.typing as npt
 import polars as pl
 from scipy import special
 
-from .._utils.array import (
+from model_diagnostics._utils._array import (
     array_name,
     get_second_dimension,
+    get_sorted_array_names,
     length_of_second_dimension,
     validate_2_arrays,
     validate_same_first_dimension,
@@ -42,7 +44,7 @@ def identification_function(
         - `"quantile"`
     level : float
         The level of the expectile of quantile. (Often called \(\alpha\).)
-        It must be `0 <= level <= 1`.
+        It must be `0 < level < 1`.
         `level=0.5` and `functional="expectile"` gives the mean.
         `level=0.5` and `functional="quantile"` gives the median.
 
@@ -77,17 +79,25 @@ def identification_function(
 
     References
     ----------
+    `[Gneiting2011]`
+
+    :   T. Gneiting.
+        "Making and Evaluating Point Forecasts”. (2011)
+        [doi:10.1198/jasa.2011.r10138](https://doi.org/10.1198/jasa.2011.r10138)
+        [arxiv:0912.0902](https://arxiv.org/abs/0912.0902)
 
     Examples
     --------
-
+    >>> identification_function(y_obs=[0, 0, 1, 1], y_pred=[-1, 1, 1 , 2])
+    array([-1,  1,  0,  1])
     """
     y_o: np.ndarray
     y_p: np.ndarray
     y_o, y_p = validate_2_arrays(y_obs, y_pred)
 
-    if functional in ("expectile", "quantile") and (level < 0 or level > 1):
-        raise ValueError(f"Argument level must fulfil 0 <= level <= 1, got {level}.")
+    if functional in ("expectile", "quantile") and (level <= 0 or level >= 1):
+        msg = f"Argument level must fulfil 0 < level < 1, got {level}."
+        raise ValueError(msg)
 
     if functional == "mean":
         return y_p - y_o
@@ -99,16 +109,17 @@ def identification_function(
         return np.greater_equal(y_p, y_o) - level
     else:
         allowed_functionals = ("mean", "median", "expectile", "quantile")
-        raise ValueError(
+        msg = (
             f"Argument functional must be one of {allowed_functionals}, got "
             f"{functional}."
         )
+        raise ValueError(msg)
 
 
 def compute_bias(
     y_obs: npt.ArrayLike,
     y_pred: npt.ArrayLike,
-    feature: Optional[npt.ArrayLike] = None,
+    feature: Optional[Union[npt.ArrayLike, pl.Series]] = None,
     weights: Optional[npt.ArrayLike] = None,
     *,
     functional: str = "mean",
@@ -148,12 +159,13 @@ def compute_bias(
         - `"quantile"`
     level : float
         The level of the expectile of quantile. (Often called \(\alpha\).)
-        It must be `0 <= level <= 1`.
+        It must be `0 < level < 1`.
         `level=0.5` and `functional="expectile"` gives the mean.
         `level=0.5` and `functional="quantile"` gives the median.
     n_bins : int
         The number of bins for numerical features and the maximal number of (most
-        frequent) categories shown for categorical features.
+        frequent) categories shown for categorical features. Due to ties, the effective
+        number of bins might be smaller than `n_bins`.
 
     Returns
     -------
@@ -193,7 +205,7 @@ def compute_bias(
     = \frac{1}{n-1} \frac{1}{\sum_i w_i} \sum_i w_i (V(m(x_i), y_i) - \bar{V})^2\) with
     the implied relation \(\operatorname{Var}(V(m(x_i), y_i)) \sim \frac{1}{w_i} \).
     If your weights are for repeated observations, so-called frequency weights, then
-    the above estimate is on the conservative site because is used \(n - 1\) instead
+    the above estimate is on the conservative side because it uses \(n - 1\) instead
     of \((\sum_i w_i) - 1\).
 
     References
@@ -202,38 +214,135 @@ def compute_bias(
 
     :   T. Fissler, C. Lorentzen, and M. Mayer.
         "Model Comparison and Calibration Assessment". (2022)
-        [arxiv:https://arxiv.org/abs/2202.12780](https://arxiv.org/abs/2202.12780).
+        [arxiv:2202.12780](https://arxiv.org/abs/2202.12780).
+
+    Examples
+    --------
+    >>> compute_bias(y_obs=[0, 0, 1, 1], y_pred=[-1, 1, 1 , 2])
+    shape: (1, 5)
+    ┌───────────┬────────────┬──────────────┬─────────────┬──────────┐
+    │ bias_mean ┆ bias_count ┆ bias_weights ┆ bias_stderr ┆ p_value  │
+    │ ---       ┆ ---        ┆ ---          ┆ ---         ┆ ---      │
+    │ f64       ┆ u32        ┆ f64          ┆ f64         ┆ f64      │
+    ╞═══════════╪════════════╪══════════════╪═════════════╪══════════╡
+    │ 0.25      ┆ 4          ┆ 4.0          ┆ 0.478714    ┆ 0.637618 │
+    └───────────┴────────────┴──────────────┴─────────────┴──────────┘
+    >>> compute_bias(y_obs=[0, 0, 1, 1], y_pred=[-1, 1, 1 , 2],
+    ... feature=["a", "a", "b", "b"])
+    shape: (2, 6)
+    ┌─────────┬───────────┬────────────┬──────────────┬─────────────┬─────────┐
+    │ feature ┆ bias_mean ┆ bias_count ┆ bias_weights ┆ bias_stderr ┆ p_value │
+    │ ---     ┆ ---       ┆ ---        ┆ ---          ┆ ---         ┆ ---     │
+    │ str     ┆ f64       ┆ u32        ┆ f64          ┆ f64         ┆ f64     │
+    ╞═════════╪═══════════╪════════════╪══════════════╪═════════════╪═════════╡
+    │ a       ┆ 0.0       ┆ 2          ┆ 2.0          ┆ 1.0         ┆ 1.0     │
+    │ b       ┆ 0.5       ┆ 2          ┆ 2.0          ┆ 0.5         ┆ 0.5     │
+    └─────────┴───────────┴────────────┴──────────────┴─────────────┴─────────┘
     """
     validate_same_first_dimension(y_obs, y_pred)
-    if feature is not None:
-        validate_same_first_dimension(y_obs, feature)
     n_pred = length_of_second_dimension(y_pred)
-    if n_pred == 0:
-        pred_names = [""]
-    else:
-        pred_names = []
-        for i in range(n_pred):
-            x = get_second_dimension(y_pred, i)
-            pred_names.append(array_name(x, default=str(i)))
+    pred_names, _ = get_sorted_array_names(y_pred)
 
     if weights is not None:
         validate_same_first_dimension(weights, y_obs)
         w = np.asarray(weights)
         if w.ndim > 1:
-            raise ValueError(
-                "The array weights must be 1-dimensional, got weights.ndim={w.ndim}."
-            )
+            msg = f"The array weights must be 1-dimensional, got weights.ndim={w.ndim}."
+            raise ValueError(msg)
     else:
-        w = np.ones_like(y_obs)
+        w = np.ones_like(y_obs, dtype=float)
 
     df_list = []
     with pl.StringCache():
+        is_categorical = False
+        is_numeric = False
+        is_string = False
+
+        if feature is None:
+            feature_name = None
+        else:
+            feature_name = array_name(feature, default="feature")
+            # The following statement, i.e. possibly the creation of a pl.Categorical,
+            # MUST be under the StringCache context manager!
+            feature = pl.Series(name=feature_name, values=feature)
+            validate_same_first_dimension(y_obs, feature)
+            if feature.dtype is pl.Categorical:
+                is_categorical = True
+            elif feature.dtype in [pl.Utf8, pl.Object]:
+                # We could convert strings to categoricals.
+                is_string = True
+            elif feature.is_float():
+                is_numeric = True
+                # We treat NaN as Null values, numpy will see a Null as a NaN.
+                if feature.is_float():
+                    feature = feature.fill_nan(None)
+            else:
+                # integers
+                is_numeric = True
+
+            if is_categorical or is_string:
+                # For categorical and string features, knowing the frequency table in
+                # advance makes life easier in order to make results consistent.
+                # Consider
+                #     feature  counts
+                #         "a"      3
+                #         "b"      2
+                #         "c"      2
+                #         "d"      1
+                # with n_bins = 2. As we want the effective number of bins to be at
+                # most n_bins, we want, in the above case, only "a" in the final
+                # result. Therefore, we need to internally decrease n_bins to 1.
+                value_count = feature.value_counts(sort=True)
+                if n_bins >= value_count.shape[0]:
+                    n_bins = value_count.shape[0]
+                else:
+                    n = value_count["counts"][n_bins]
+                    n_bins_tmp = value_count.filter(pl.col("counts") >= n).shape[0]
+                    if n_bins_tmp > n_bins:
+                        n_bins = value_count.filter(pl.col("counts") > n).shape[0]
+                    else:
+                        n_bins = n_bins_tmp
+                if n_bins == 0:
+                    msg = (
+                        "Due to ties, the effective number of bins is 0. "
+                        f"Consider to increase n_bins>={n_bins_tmp}."
+                    )
+                    warnings.warn(msg, UserWarning)
+            else:
+                # Binning
+                # We use method="inverted_cdf" (same as "lower") instead of the
+                # default "linear" because "linear" produces as many unique values
+                # as before.
+                q = np.nanquantile(
+                    feature,
+                    # Improved rounding errors by using integers inside linspace.
+                    q=np.linspace(0 + 1, n_bins - 1, num=n_bins - 1) / n_bins,
+                    method="inverted_cdf",
+                )  # type: ignore
+                q = np.unique(q)  # Some quantiles might be the same.
+                # We want: bins[i-1] < x <= bins[i]
+                f_binned = np.digitize(feature, bins=q, right=True)  # type: ignore
+                # Now, we insert Null values again at the original places.
+                f_binned = (
+                    pl.LazyFrame(
+                        [
+                            pl.Series("__f_binned", f_binned, dtype=feature.dtype),
+                            feature,
+                        ]
+                    )
+                    .select(
+                        pl.when(pl.col(feature_name).is_null())
+                        .then(None)
+                        .otherwise(pl.col("__f_binned"))
+                        .alias("bin")
+                    )
+                    .collect()
+                    .get_column("bin")
+                )
+
         for i in range(len(pred_names)):
             # Loop over columns of y_pred.
-            if n_pred == 0:
-                x = y_pred
-            else:
-                x = get_second_dimension(y_pred, i)
+            x = y_pred if n_pred == 0 else get_second_dimension(y_pred, i)
 
             bias = identification_function(
                 y_obs=y_obs,
@@ -243,7 +352,6 @@ def compute_bias(
             )
 
             if feature is None:
-                feature_name = None
                 bias_mean = np.average(bias, weights=w)
                 bias_weights = np.sum(w)
                 bias_count = bias.shape[0]
@@ -254,13 +362,12 @@ def compute_bias(
                 df = pl.DataFrame(
                     {
                         "bias_mean": [bias_mean],
-                        "bias_count": [bias_count],
+                        "bias_count": pl.Series([bias_count], dtype=pl.UInt32),
                         "bias_weights": [bias_weights],
                         "bias_stderr": [np.sqrt(bias_stddev)],
                     }
                 )
             else:
-                feature_name = array_name(feature, default="feature")
                 df = pl.DataFrame(
                     {
                         "y_obs": y_obs,
@@ -271,17 +378,6 @@ def compute_bias(
                     }
                 )
 
-                is_categorical = False
-                is_numeric = False
-                is_string = False
-                if df.get_column(feature_name).dtype is pl.Categorical:
-                    is_categorical = True
-                elif df.get_column(feature_name).dtype in [pl.Utf8, pl.Object]:
-                    # We could convert strings to categoricals.
-                    is_string = True
-                else:
-                    is_numeric = True
-
                 agg_list = [
                     pl.col("bias_mean").first(),
                     pl.count("bias").alias("bias_count"),
@@ -298,21 +394,8 @@ def compute_bias(
                 if is_categorical or is_string:
                     groupby_name = feature_name
                 else:
-                    # Binning
-                    # We use method="inverted_cdf" (same as "lower") instead of the
-                    # default "linear" because "linear" produces as many unique values
-                    # as before.
-                    q = np.quantile(
-                        feature,
-                        # q=np.linspace(0 + 1 / n_bins, 1 - 1 / n_bins, num=n_bins - 1),
-                        # improved rounding errors
-                        q=np.linspace(0 + 1, n_bins - 1, num=n_bins - 1) / n_bins,
-                        method="inverted_cdf",
-                    )  # type: ignore
-                    q = np.unique(q)  # Some quantiles might be the same.
-                    # bins[i-1] < x <= bins[i]
-                    f_binned = np.digitize(feature, bins=q, right=True)  # type: ignore
-                    df = df.hstack([pl.Series("bin", f_binned)])
+                    # See above for the creation of the binned feature f_binned.
+                    df = df.hstack([f_binned])
                     groupby_name = "bin"
                     agg_list.append(pl.col(feature_name).mean())
 
@@ -340,9 +423,9 @@ def compute_bias(
                             .alias("bias_stderr"),
                         ]
                     )
-                    .sort("bias_count", reverse=True)
+                    .sort("bias_count", descending=True)
                     .head(n_bins)
-                    .sort(feature_name)
+                    .sort(feature_name, descending=False)
                 )
                 if is_numeric:
                     df = df.drop("bin")
@@ -386,10 +469,7 @@ def compute_bias(
 
             # Add column "model".
             if n_pred > 0:
-                if feature_name == "model":
-                    model_col_name = "model_"
-                else:
-                    model_col_name = "model"
+                model_col_name = "model_" if feature_name == "model" else "model"
                 df = df.with_columns(
                     pl.Series(model_col_name, [pred_names[i]] * df.shape[0])
                 )
@@ -400,7 +480,13 @@ def compute_bias(
                 col_selection.append(model_col_name)
             if feature_name is not None and feature_name in df.columns:
                 col_selection.append(feature_name)
-            col_selection += ["bias_mean", "bias_count", "bias_stderr", "p_value"]
+            col_selection += [
+                "bias_mean",
+                "bias_count",
+                "bias_weights",
+                "bias_stderr",
+                "p_value",
+            ]
             df_list.append(df.select(col_selection))
 
         df = pl.concat(df_list)
