@@ -6,53 +6,14 @@ import polars as pl
 import pyarrow as pa
 import pytest
 from polars.testing import assert_frame_equal
-from scipy.optimize import root_scalar
-from scipy.stats import ttest_1samp
+from scipy.special import stdtr
+from scipy.stats import expectile, ttest_1samp
 
 from model_diagnostics.calibration import compute_bias, identification_function
 
 
-# Note: expectile might arrive with scipy 1.10.
-# Backport of scipy.stats.expectile
-# https://github.com/scipy/scipy/pull/17039
-def expectile(a, alpha=0.5, *, dtype=None, weights=None):  # pragma: no cover
-    r"""Compute the expectile along the specified axis."""
-    if alpha < 0 or alpha > 1:
-        raise ValueError("The expectile level alpha must be in the range [0, 1].")
-    a = np.asarray(a, dtype=dtype)
-
-    if weights is not None:
-        weights = np.asarray(weights, dtype=dtype)
-
-    # This is the empirical equivalent of Eq. (13) with identification
-    # function from Table 9 (omitting a factor of 2) in
-    # Gneiting, T. (2009). "Making and Evaluating Point Forecasts".
-    # Journal of the American Statistical Association, 106, 746 - 762.
-    # https://doi.org/10.48550/arXiv.0912.0902
-    def first_order(t):
-        return alpha * np.average(np.fmax(0, a - t), weights=weights) - (
-            1 - alpha
-        ) * np.average(np.fmax(0, t - a), weights=weights)
-
-    if alpha >= 0.5:
-        x0 = np.average(a, weights=weights)
-        x1 = np.amax(a)
-    else:
-        x1 = np.average(a, weights=weights)
-        x0 = np.amin(a)
-
-    if x0 == x1:
-        # a has a single unique element
-        return x0
-
-    # Note that the expectile is the unique solution, so no worries about
-    # finding a wrong root.
-    res = root_scalar(first_order, x0=x0, x1=x1)
-    return res.root
-
-
 @pytest.mark.parametrize(
-    "functional, est",
+    ("functional", "est"),
     [
         ("mean", np.mean),
         ("median", np.median),
@@ -75,10 +36,12 @@ def test_identification_function_equal_input(functional, est, level):
 
 
 @pytest.mark.parametrize(
-    "functional, level, msg",
+    ("functional", "level", "msg"),
     [
         ("no good functional", 0.5, "Argument functional must be one of"),
-        ("quantile", 1.1, "Argument level must fulfil 0 <= level <= 1"),
+        ("quantile", 0, "Argument level must fulfil 0 < level < 1"),
+        ("quantile", 1, "Argument level must fulfil 0 < level < 1"),
+        ("quantile", 1.1, "Argument level must fulfil 0 < level < 1"),
     ],
 )
 def test_identification_function_raises(functional, level, msg):
@@ -88,7 +51,7 @@ def test_identification_function_raises(functional, level, msg):
 
 
 @pytest.mark.parametrize(
-    "feature, f_grouped",
+    ("feature", "f_grouped"),
     [
         (
             pa.DictionaryArray.from_arrays([0, 0, 1, 1, 1], ["b", "a"]),
@@ -102,6 +65,14 @@ def test_identification_function_raises(functional, level, msg):
             pa.array([0.1, 0.1, 0.9, 0.9, 0.9]),
             pa.array([0.1, 0.9]),
         ),
+        (
+            pa.array([None, np.nan, 1.0, 1, 1]),
+            pa.array([None, 1.0]),
+        ),
+        (
+            pa.array(["a", "a", None, None, None]),
+            pa.array(["a", None]),
+        ),
     ],
 )
 def test_compute_bias(feature, f_grouped):
@@ -117,6 +88,7 @@ def test_compute_bias(feature, f_grouped):
                 "feature": feature,
             }
         )
+        # V = [1, 0, 0, -2, -1]
         df_bias = compute_bias(
             y_obs=df.column("y_obs"),
             y_pred=df.column("y_pred"),
@@ -127,6 +99,7 @@ def test_compute_bias(feature, f_grouped):
                 "feature": f_grouped,
                 "bias_mean": [0.5, -1],
                 "bias_count": pl.Series(values=[2, 3], dtype=pl.UInt32),
+                "bias_weights": pl.Series(values=[2, 3], dtype=pl.Float64),
                 "bias_stderr": np.sqrt([(0.25 + 0.25), (1 + 1 + 0) / 2])
                 / np.sqrt([2, 3]),
                 "p_value": [
@@ -134,7 +107,33 @@ def test_compute_bias(feature, f_grouped):
                     ttest_1samp([0, -2, -1], 0).pvalue,
                 ],
             }
+        ).sort("feature")
+        assert_frame_equal(df_bias, df_expected, check_exact=False)
+
+        # Same with weights.
+        feature = pl.Series(values=feature)
+        df_bias = compute_bias(
+            # y_obs=[0.5, (1 * 2 + 0.5 * 4) / 1.5, (0.5 * 4 + 3) / 1.5]
+            y_obs=[0.5, 8 / 3, 10 / 3],
+            y_pred=[1, 2, 2],
+            weights=[2, 1.5, 1.5],
+            feature=feature.take([0, 4, 4]).alias("feature"),
         )
+        # V = [0.5, -2/3, -4/3]
+        df_expected = pl.DataFrame(
+            {
+                "feature": f_grouped,
+                "bias_mean": [0.5, -1],
+                "bias_count": pl.Series(values=[1, 2], dtype=pl.UInt32),
+                "bias_weights": pl.Series(values=[2, 3], dtype=pl.Float64),
+                # For feature[0], the variance is 0 because there is only one
+                # observation. For feature[4], a direct calculation gives:
+                # SE = sqrt((1.5 * 1/9 + 1.5 * 1/9) / 3 / (2-1)) = sqrt(1 / (3 * 3 * 1))
+                #    = sqrt(1/9) = 1/3
+                "bias_stderr": [0.0, 1 / 3],
+                "p_value": [np.nan, 2 * stdtr(2 - 1, -np.abs(-1 / (1 / 3)))],
+            }
+        ).sort("feature")
         assert_frame_equal(df_bias, df_expected, check_exact=False)
 
 
@@ -146,6 +145,7 @@ def test_compute_bias_feature_none():
             "y_pred": [1, 1, 2, 2, 2],
         }
     )
+    # V = [1, 0, 0, -2, -1]
     df_bias = compute_bias(
         y_obs=df.column("y_obs"),
         y_pred=df.column("y_pred"),
@@ -154,7 +154,8 @@ def test_compute_bias_feature_none():
     df_expected = pl.DataFrame(
         {
             "bias_mean": [-0.4],  # (1 + 0 + 0 - 2 - 1) / 5
-            "bias_count": [5],
+            "bias_count": pl.Series(values=[5], dtype=pl.UInt32),
+            "bias_weights": [5.0],
             "bias_stderr": [np.std([1, 0, 0, -2, -1], ddof=1) / np.sqrt(5)],
             "p_value": [
                 ttest_1samp([1, 0, 0, -2, -1], 0).pvalue,
@@ -163,6 +164,30 @@ def test_compute_bias_feature_none():
     )
     assert_frame_equal(df_bias, df_expected)
 
+    # Same with weights.
+    df_bias = compute_bias(
+        # y_obs=[0.5, (1 * 2 + 0.5 * 4) / 1.5, (0.5 * 4 + 3) / 1.5]
+        y_obs=[0.5, 8 / 3, 10 / 3],
+        y_pred=[1, 2, 2],
+        weights=[2, 1.5, 1.5],
+        feature=None,
+    )
+    # V = [0.5, -2/3, -4/3]
+    df_expected = (
+        df_expected.replace("bias_count", pl.Series(values=[3], dtype=pl.UInt32))
+        .replace(
+            # SE = sqrt((2 * 0.9**2 + 1.5 * 0.8**2/9 + 1.5 * 2.8**2/9) / 5 / (3-1))
+            #    = sqrt(91 / 300)
+            "bias_stderr",
+            pl.Series(values=[np.sqrt(91 / 300)]),
+        )
+        .replace(
+            "p_value",
+            pl.Series(values=[2 * stdtr(3 - 1, -np.abs(-0.4 / np.sqrt(91 / 300)))]),
+        )
+    )
+    assert_frame_equal(df_bias, df_expected, check_exact=False)
+
 
 def test_compute_bias_numerical_feature():
     """Test compute_bias for a numerical feature."""
@@ -188,6 +213,7 @@ def test_compute_bias_numerical_feature():
             "feature": 0.045 + 0.1 * np.arange(10),
             "bias_mean": 0.955 - 0.1 * np.arange(10),
             "bias_count": n_steps * np.ones(n_bins, dtype=np.uint32),
+            "bias_weights": [n_obs / n_bins] * n_bins,
             "bias_stderr": [
                 np.std(bias[n : n + n_steps], ddof=1) / np.sqrt(n_steps)
                 for n in range(0, n_obs, n_steps)
@@ -230,7 +256,8 @@ def test_compute_bias_multiple_predictions(feature_type):
                 "model": ["model_1", "model_1", "model_2", "model_2"],
                 "nice_feature": f_expected,
                 "bias_mean": [0.0, -1, 2, 1],
-                "bias_count": np.array([5] * 4, dtype=np.uint32),
+                "bias_count": pl.Series(values=[5] * 4, dtype=pl.UInt32),
+                "bias_weights": [5.0] * 4,
                 "bias_stderr": [0.0] * 4,
                 "p_value": [np.nan] * 4,
             }
@@ -274,6 +301,79 @@ def test_compute_bias_multiple_predictions(feature_type):
         df_expected = df_expected.replace(
             "model_", pl.Series(["model_1", "model_1", "model_2", "model_2"])
         )
-        print(f"left\n{df_bias}")
-        print(f"right\n{df_expected}")
         assert_frame_equal(df_bias, df_expected, check_exact=False)
+
+
+@pytest.mark.parametrize("feature", [None, [1, 1]])
+def test_compute_bias_model_name_order(feature):
+    """Test that the order of the colum model is not sorted."""
+    y_obs = [0, 0]
+    y_pred = pl.DataFrame(
+        {
+            "model_3": [1, 1],
+            "model_1": [0, 1],
+            "model_2": [0, 1],
+        }
+    )
+    df_bias = compute_bias(
+        y_obs=y_obs,
+        y_pred=y_pred,
+        feature=feature,
+    )
+    assert df_bias.get_column("model").to_list() == ["model_3", "model_1", "model_2"]
+
+
+def test_compute_bias_many_sparse_feature_values():
+    """Test that compute_bias returns same values for high cardinality feature."""
+    n_obs = 100
+    y_obs = np.arange(n_obs)
+    y_pred = pl.DataFrame(
+        {
+            "model_1": np.arange(n_obs) + 0.5,
+            "model_2": (y_obs - 5) ** 2,
+            "model_3": (y_obs - 3) ** 2,
+        }
+    )
+    rng = np.random.default_rng(42)
+    feature = rng.integers(low=0, high=n_obs // 2, size=n_obs).astype(str)
+
+    df = compute_bias(
+        y_obs=y_obs,
+        y_pred=y_pred,
+        feature=feature,
+        n_bins=10,
+    )
+
+    assert set(df.filter(pl.col("model") == "model_1")["feature"]) == set(
+        df.filter(pl.col("model") == "model_2")["feature"]
+    )
+    assert set(df.filter(pl.col("model") == "model_1")["feature"]) == set(
+        df.filter(pl.col("model") == "model_3")["feature"]
+    )
+
+
+def test_compute_bias_warning_for_n_bins():
+    """Test that compute_bias gives warning for n_bins to small."""
+    y_obs = np.arange(6)
+    y_pred = y_obs + 1
+    feature = ["a", "a", "b", "b", "c", "c"]
+
+    with pytest.warns(
+        UserWarning, match="Due to ties, the effective number of bins is 0"
+    ):
+        df = compute_bias(
+            y_obs=y_obs,
+            y_pred=y_pred,
+            feature=feature,
+            n_bins=2,
+        )
+
+    assert df.shape[0] == 0
+
+
+def test_compute_bias_raises_weights_shape():
+    y_obs, y_pred = np.arange(5), np.arange(5)
+    weights = np.arange(5)[:, None]
+    msg = "The array weights must be 1-dimensional, got weights.ndim=2."
+    with pytest.raises(ValueError, match=msg):
+        compute_bias(y_obs, y_pred, weights=weights)
